import pyxel
import math

## global vars

PLAYER_WIDTH = 3
PLAYER_HEIGHT = 3
PLAYER_SPEED = 2

BULLET_COLOR = 7
BULLET_SPEED = 2

ENEMY_WIDTH = 3
ENEMY_HEIGHT = 3
ENEMY_COLOR = [3, 11, 6]

MOUSE_WIDTH = 1
MOUSE_HEIGHT = 1

## global lists

enemies = []
bullets = []
blasts = []

def update_list(list):
    for elem in list:
        elem.update()

def draw_list(list):
    for elem in list:
        elem.draw()

def cleanup_list(list):
    i = 0
    while i < len(list):
        elem = list[i]
        if not elem.is_alive:
            list.pop(i)
        else:
            i += 1

class App:
    def __init__(self):
        pyxel.init(300, 200, title="Zerobit's game1", fps=60)
        self.score = 0
        self.player = Player(round(pyxel.width / 2),110)
        self.mouse = Mouse()
        pyxel.mouse(visible=False)
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btnp(pyxel.KEY_Q):
            pyxel.quit()
        self.update_field()
        self.player.update()
        self.mouse.update()
        update_list(bullets)
        cleanup_list(bullets)
        update_list(enemies)
        cleanup_list(enemies)

    def update_field(self):
        if pyxel.frame_count % 6 == 0:
            enemies.append(Enemy(self.player.x, self.player.y))

    def draw(self):
        pyxel.cls(0)
        self.player.draw()
        draw_list(bullets)
        draw_list(enemies)
        self.mouse.draw()

class Enemy:
    def __init__(self,player_x, player_y):
        self.init_x = pyxel.rndi(0, pyxel.width)
        self.init_y = pyxel.rndi(0, pyxel.height)
        self.min_distance = 30
        self.x = 0
        self.y = 0
        self.init_player_x = player_x
        self.init_player_y = player_y
        self.w = ENEMY_WIDTH
        self.h = ENEMY_HEIGHT
        self.is_alive = True
        self.spawn()
        enemies.append(self)
    
    def spawn(self):
        #spawn bad guy, also checks players location and rerolls spawn if too close
        while True:
            #check distance to player
            distance_player = self.check_distance((self.init_x, self.init_y),(self.init_player_x, self.init_player_y))
            if distance_player < self.min_distance:
                self.init_x = pyxel.rndi(0, pyxel.width)
                self.init_y = pyxel.rndi(0, pyxel.height)
                continue
            else:
                for mob in enemies:
                        distance_enemy = self.check_distance((self.init_x, self.init_y),(mob.init_x, mob.init_y))
                        
                        if distance_enemy < self.min_distance:
                            #reroll the spawn
                            self.init_x = pyxel.rndi(0, pyxel.width)
                            self.init_y = pyxel.rndi(0, pyxel.height)
                            continue
                        else:
                            break
                break
        self.x, self.y = self.init_x, self.init_y

    @staticmethod
    def check_distance(point1: tuple[int,int], point2:tuple[int,int]):
        return math.sqrt((point1[0] - point2[0])**2 + (point1[1] - point2[1])**2)

    def update(self):
        print(f"enemy count:{len(enemies)} frame:{pyxel.frame_count} bullets:{len(bullets)}")

    def draw(self):
        pyxel.rect(self.x, self.y, self.w, self.h, ENEMY_COLOR[pyxel.rndi(0,2)])


class Bullet:
    def __init__(self, x, y, mouse_x, mouse_y):
        self.x = round(x + 1)
        self.y = round(y + 1)
        self.target_x = round(mouse_x)
        self.target_y = round(mouse_y)
        self.is_alive = True 
        self.bullet_path = []
        self.target_line()
        bullets.append(self)
    
    def calc_line(self, x1, y1, x2, y2, distance):
        # Calculate the direction vector
        delta_x = x2 - x1
        delta_y = y2 - y1
        # Normalize the direction vector
        mag = math.sqrt(delta_x**2 + delta_y**2)
        delta_x /= mag
        delta_y /= mag
        # Extrapolate the point
        px = int(x1 + delta_x * distance)
        py = int(y1 + delta_y * distance)
        return px, py

    def target_line(self):
        if self.x == self.target_x and self.y < self.target_y:
            #under player
            y_path = list(range(self.y, pyxel.height))
            for point in y_path:
                line = (self.x, point)
                self.bullet_path.append(line)
        if self.x == self.target_x and self.y > self.target_y:
            #above player
            y_path = list(range(0, self.y))
            for point in y_path:
                line = (self.x, point)
                self.bullet_path.insert(0, line)

        if self.x > self.target_x:
            print("left side")
            for point in range(pyxel.width):
                self.bullet_path.append(self.calc_line(self.x, self.y,self.target_x, self.target_y, point))

        if self.x < self.target_x:
            print("right side")
            for point in range(pyxel.width):
                self.bullet_path.append(self.calc_line(self.x, self.y,self.target_x, self.target_y, point))

    def update(self):
        if len(self.bullet_path) == 0:
                self.is_alive = False
        else:
            if self.bullet_path[0][0] >= pyxel.width or self.bullet_path[0][1] >= pyxel.height:
                self.is_alive = False
            if self.bullet_path[0][0] <= 0 or self.bullet_path[0][1] <= 0:
                self.is_alive = False
            self.bullet_path.pop(0)


    def draw(self):
        if self.is_alive == True and len(self.bullet_path) !=0:
            pyxel.pset(self.bullet_path[0][0], self.bullet_path[0][1], BULLET_COLOR)

class Mouse:
    def __init___(self):
        self.x = 0
        self.y = 0
        self.lb = False
        self.rb = False

    def update(self):
        self.x = pyxel.mouse_x
        self.y = pyxel.mouse_y
        self.lb = pyxel.btn(pyxel.MOUSE_BUTTON_LEFT)
        self.rb = pyxel.btn(pyxel.MOUSE_BUTTON_RIGHT)
        #keeping mouse in bounds
        self.x = max(self.x, 0)
        self.x = min(self.x, pyxel.width - MOUSE_WIDTH)
        self.y = max(self.y, 0)
        self.y = min(self.y, pyxel.height - MOUSE_HEIGHT)
    
    def draw(self):
        #this makes a cross for the mouse cursor
        pyxel.line(self.x-2, self.y, self.x+2, self.y, 8)
        pyxel.line(self.x, self.y-2, self.x, self.y+2, 8)

class Player:
    def __init__(self, x, y):
        self.x = round(x)
        self.y = round(y)
        self.w = PLAYER_WIDTH
        self.h = PLAYER_HEIGHT
        self.is_alive = True
        enemies.append(self)
    
    def update(self):
        if pyxel.btn(pyxel.KEY_UP) or pyxel.btn(pyxel.KEY_W):
            self.y = (self.y - 1)
        if pyxel.btn(pyxel.KEY_DOWN) or pyxel.btn(pyxel.KEY_S):
            self.y = (self.y + 1)
        if pyxel.btn(pyxel.KEY_RIGHT) or pyxel.btn(pyxel.KEY_D):
            self.x = (self.x + 1)
        if pyxel.btn(pyxel.KEY_LEFT) or pyxel.btn(pyxel.KEY_A):
            self.x = (self.x - 1)
        if pyxel.btn(pyxel.MOUSE_BUTTON_LEFT):
            Bullet(
                self.x,
                self.y,
                pyxel.mouse_x,
                pyxel.mouse_y
                )
        #bounds for the player
        self.x = max(self.x, 0)
        self.x = min(self.x, pyxel.width - PLAYER_WIDTH)
        self.y = max(self.y, 0)
        self.y = min(self.y, pyxel.height - PLAYER_HEIGHT)


    def draw(self):
        pyxel.rect(self.x, self.y, 3, 3, 5)

App()